# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common.db import models


class CouchDBSchema(models.DatastoreSchema):
    """Represents a CouchDB schema and its associated properties."""

    @property
    def _max_schema_name_length(self):
        return 32


class CouchDBUser(models.DatastoreUser):
    """Represents a CouchDB user and its associated properties."""

    @property
    def schema_model(self):
        return CouchDBSchema
