.. -*- rst -*-

============
API versions
============

Lists information for all Database Service API versions.


List versions
~~~~~~~~~~~~~

.. rest_method::  GET /

Lists information about all Database Service API versions. No authentication
needed.

Normal response codes: 200

Response Example
----------------

.. literalinclude:: samples/versions-response.json
   :language: javascript
